from __future__ import annotations

from httpx import request as _request

__all__ = ('get_response', 'DEFAULT_REQUEST_KWARGS')

DEFAULT_REQUEST_KWARGS = {'headers': {'user-agent': 'Mozilla/5.0'}}


def get_response(*args, **kwargs):
    for k, v in DEFAULT_REQUEST_KWARGS.items():
        kwargs.setdefault(k, v)
    return _request(*args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Iterator, Sequence, SupportsIndex, TypeVar

from .._bs4 import BeautifulSoup
from ..api import get_response
from ..errors import EmptyPage
from ..stream.player import *

__all__ = ('Page', 'PageNumber', 'InlineItem')

PageNumber = TypeVar('PageNumber', Iterable[int], slice, int)
T = TypeVar('T')


def _range_from_slice(obj: slice | T) -> range | T:
    if isinstance(obj, slice):
        return range(*[i for i in (obj.start, obj.stop, obj.step) if i is not None])
    return obj


@dataclass(frozen=True)
class InlineItem:
    """Content Inline Item view"""
    __slots__ = ('url', 'name', 'info', 'poster')
    url: str
    name: str
    info: str
    poster: str  # image url

    @property
    def player(self) -> PlayerType:
        return Player(self.url)


class Page:
    """Ajax class for HDRezka search"""
    __slots__ = ('_page', '_page_format')

    def __init__(self, url: str = 'https://rezka.ag/'):
        self.page = url

    def __iter__(self) -> Iterator[InlineItem]:
        """
        Returns the generator of all found articles
        """
        page = 1
        while True:
            try:
                yield from self.page_iter(page)
            except EmptyPage(page):
                return
            page += 1

    @property
    def page(self) -> str:
        return self._page

    @page.setter
    def page(self, value):
        if not isinstance(value, str):
            value = str(value)
        if not value.startswith('https://'):
            value = f'https://{value}'
        self._page = value
        self._page_format = self._concat_paginator(value.removesuffix('/')).format

    @staticmethod
    def _concat_paginator(url: str):
        return f'{url}/page/{{0}}'

    def page_iter(self, page: PageNumber = 1) -> Iterator[InlineItem] | None:
        page = _range_from_slice(page)
        is_not_iter = not isinstance(page, Iterable)
        if is_not_iter:
            page = page,
        for page in page:
            items = BeautifulSoup(get_response('GET', self._page_format(page)).text
                                  ).find_all(class_='b-content__inline_item')
            if not items:
                if is_not_iter:
                    raise EmptyPage(page)
                return
            yield from (InlineItem(
                (a := (link := i.find(class_='b-content__inline_item-link')).find('a'))['href'],
                a.text,
                link.find('div').text,
                i.find(class_='b-content__inline_item-cover').find('img').get('src', ''))
                for i in items)

    def get_pages(self, page: PageNumber = 1) -> tuple[InlineItem]:
        return *self.page_iter(page),

    def __getitem__(self, item: PageNumber | Sequence[PageNumber | SupportsIndex | slice]
                    ) -> tuple[InlineItem] | InlineItem:
        if isinstance(item, Sequence) and len(item) == 2:
            page, index = item
            return self.get_pages(page)[index]
        return self.get_pages(item)

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.page!r})'



from __future__ import annotations

from collections import defaultdict
from functools import lru_cache
from typing import Any, SupportsInt, TypeVar

from .._bs4 import BeautifulSoup
from ..api.ajax import Ajax
from ..errors import UnknownContentType
from ..post import *

__all__ = ('Player', 'PlayerType')


class PlayerBase:
    __slots__ = ('post',)

    def __init__(self, url_or_cast: Any):
        if isinstance(url_or_cast, PlayerBase):
            self.post = url_or_cast.post
            return
        elif not isinstance(url_or_cast, str):
            url_or_cast = str(url_or_cast)
        self.post = Post(url_or_cast)

    def _translator(self, translator_id: SupportsInt = None):
        if translator_id is None:
            return self.post.translator_id
        translator_id = int(translator_id)
        return self.post.translators.ids[abs(translator_id)] if translator_id <= 0 else translator_id

    def __repr__(self):
        return f"{self.__class__.__qualname__}({short_url(self.post.url)!r})"


class PlayerMovie(PlayerBase):
    __slots__ = ()

    def get_stream(self, translator_id: SupportsInt = None) -> URLs:
        return URLs(Ajax.get_movie(self.post.id, self._translator(translator_id))['url'])


class PlayerSeries(PlayerBase):
    __slots__ = ()

    def get_episodes(self, translator_id: SupportsInt = None) -> defaultdict[int, tuple[int]]:
        episodes = BeautifulSoup(Ajax.get_episodes(self.post.id, self._translator(translator_id))['episodes'])
        result: defaultdict[int, tuple[int]] = defaultdict(tuple)
        for i in episodes.find_all(class_='b-simple_episode__item', attrs=('data-season_id', 'data-episode_id')):
            result[int(i.get('data-season_id'))] += int(i.get('data-episode_id')),
        return result

    def get_stream(self, season: int, episode: int, translator_id: SupportsInt = None) -> URLs:
        return URLs(Ajax.get_stream(
            self.post.id,
            self._translator(translator_id),
            season, episode)['url'])


PlayerType = TypeVar('PlayerType', PlayerBase, PlayerMovie, PlayerSeries)


@lru_cache(512)
def player(url_or_path: Any) -> PlayerType:
    """
    Returns either Player Series if series, or PlayerMovie if movie, otherwise raises UnknownContentType
    """
    p = PlayerBase(url_or_path)
    type = p.post.type
    if type == 'tv_series':
        return PlayerSeries(p)
    if type == 'movie':
        return PlayerMovie(p)
    raise UnknownContentType(type)


Player = player



from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple

__all__ = ('Rating', 'Place', 'Release', 'AgeRating', 'Duration', 'Poster')


@dataclass(frozen=True)
class Rating:
    __slots__ = ('service', 'rating', 'votes')
    service: str
    rating: int | float
    votes: int


@dataclass(frozen=True)
class Place:
    __slots__ = ('name', 'place')
    name: str
    place: int


@dataclass(frozen=True)
class Release:
    __slots__ = ('day', 'year')
    year: int
    day: str


@dataclass(frozen=True)
class AgeRating:
    __slots__ = ('age', 'description')
    age: int
    description: str


@dataclass(frozen=True)
class Duration:
    __slots__ = ('number', 'units')
    number: int
    units: str


class Poster(NamedTuple):
    full: str
    preview: str



from __future__ import annotations

__all__ = ('Translators',)


class Translators:
    __slots__ = ('names', 'ids', 'name_id', 'id_name')

    def __init__(self, name_id: dict[str, int]):
        self.names = *name_id,
        self.ids = *name_id.values(),
        self.id_name: dict[int, str] = {v: k for k, v in name_id.items()}
        self.name_id = name_id

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.name_id})'



from __future__ import annotations

import re
from functools import lru_cache
from typing import Iterable, SupportsInt

from .._antiobfuscation import clear_trash

__all__ = ('short_url', 'long_url', 'URLs', 'Quality', 'URL')

_findall_qualities = re.compile(r'\[([^]]+)](\S+)(?:\sor\s|$)').findall
_match_quality_int = re.compile(r'(\d+)[pi]\s*($|\w+)').match
_shorten_url_match = re.compile(r'(?:(?:https?://)?rezka\.ag/)?\D*(\d+)\S*(?:\.html)?/?', re.I).match


@lru_cache(1024)
def short_url(url: str) -> str:
    """
    Returns string rezka.ag post with format "{id}-{id}" (valid path)
    """
    parts = _shorten_url_match(url)
    if parts is None:
        return url
    id = parts.group(1)
    return f'{id}-{id}'


@lru_cache(1024)
def long_url(url: str) -> str:
    """
    Returns full url of rezka.ag post
    """
    return f'https://rezka.ag/{short_url(url)}.html'


class Quality(str):
    __slots__ = ('_i', 'addon')
    addon: str  # can contain 'ultra'

    def __init__(self, *_, **__):
        _i = _match_quality_int(self)
        if not _i:
            raise ValueError(f'{self!r} is not quality.')
        _i, self.addon = _i.groups()
        self.addon = self.addon.casefold()
        self._i = int(_i)

    def __int__(self):
        """
        returns pixels height
        """
        return self._i

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return super().__le__(other)
        if not (self.addon or other.addon):
            return self._i < int(other)
        return False


class URL(str):
    __slots__ = ('mp4',)

    def __init__(self, *_, **__):
        self.mp4 = self.removesuffix('8').removesuffix(':hls:manifest.m3u')


class URLs:
    def __init__(self, data: str | dict):
        self._data: dict[Quality, URL] = (
            {Quality(q): URL(u) for q, u in _findall_qualities(clear_trash(data))}
            if isinstance(data, str) else data)
        self.qualities: tuple[Quality] = *sorted(self._data),
        self.min = int(self.qualities[0]) if self.qualities else 1

    @property
    def best_url(self) -> URL:
        return [*self[-1]._data.values()][0]

    def __getitem__(self, item: (SupportsInt | str) | (slice | Iterable)):
        isiter = False
        if not isinstance(item, str):
            if isinstance(item, slice):
                item = item.start, item.stop, item.step
            isiter = not not isinstance(item, Iterable) and item
        if isiter:
            result = self
            for part in item:
                if part is not None:
                    result = result[part]
            result = result._data
        elif isinstance(item, int):
            if item < self.min:
                item = self.qualities[item]
                result = {item: self._data[item]}
            else:
                result = {q: v for q, v in self._data.items() if int(q) == item}
        elif isinstance(item, str):
            item = item.casefold()
            result = {q: v for q, v in self._data.items() if q.addon == item}
        else:
            raise TypeError(f'Invalid type {type(item)}')
        return self.__class__(result)

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self._data})"



class HDRezkaError(Exception):
    """Any HDRezka exception"""
    __slots__ = ()


class UnknownContentType(HDRezkaError, TypeError):
    """Invalid content type"""
    __slots__ = ()


class AjaxFail(HDRezkaError):
    """No success response"""
    __slots__ = ()


class EmptyPage(HDRezkaError):
    """Asked page is empty (see Page class)"""
    __slots__ = ()



from functools import partial

from bs4 import BeautifulSoup, FeatureNotFound

__all__ = ('BeautifulSoup',)
try:
    BeautifulSoup(features='lxml')
    BeautifulSoup = partial(BeautifulSoup, features='lxml')
except FeatureNotFound:
    BeautifulSoup = partial(BeautifulSoup, features='html.parser')



from __future__ import annotations

from .info import PostInfo
from .urls import long_url, short_url
from .._bs4 import BeautifulSoup
from ..api import get_response
from ..translators import Translators

__all__ = ('Post',)


class Post:
    """Stores information about the post"""
    __slots__ = ('url', 'translator_id', 'id', 'name', 'type', 'info', 'translators', 'other_parts_urls',
                 '_soup_inst')

    def __init__(self, url: str):
        url = long_url(url)
        _response = get_response('GET', url)
        if _response.status_code in {301, 302}:  # redirect
            _response = get_response('GET', _response.url.join(_response.headers['location']))
        _response = _response.text
        self.url = url
        self._soup_inst = BeautifulSoup(_response)
        self.type = self._get_type()

        self.translator_id = int(i) if (i := _response.split(
            'sof.tv.' + {'tv_series': 'initCDNSeriesEvents', 'movie': 'initCDNMoviesEvents'}[self.type],
            1)[-1].split(',')[1].strip()).isnumeric() else None
        self.info = self._get_post_info()
        self.translators = self._get_translators()

        self.id = self._extract_id()
        self.name = self._get_name()
        self.other_parts_urls = self._parts_urls

    def _extract_id(self):
        return int(self._soup_inst.find(id='post_id')['value'])

    def _get_name(self):
        return self._soup_inst.find(class_='b-post__title').get_text().strip()

    def _get_type(self):
        return self._soup_inst.find('meta', property='og:type')['content'].removeprefix('video.')

    def _get_post_info(self) -> PostInfo:
        return PostInfo(self._soup_inst)

    def _get_translators(self):
        translators_list = self._soup_inst.find(id='translators-list')
        arr = {child.text.strip(): int(child['data-translator_id']) for child in
               translators_list.find_all(recursive=False) if child.text} if translators_list else {}
        if not arr:
            arr[self.info.translator] = self.translator_id
        return Translators(arr)

    @property
    def _parts_urls(self) -> tuple[str]:
        self.other_parts_urls = *(
            i.attrs['data-url'] for i in
            self._soup_inst.select('.b-post__partcontent_item[data-url]')),
        return *map(short_url, self.other_parts_urls),

    def __repr__(self):
        return f'{self.__class__.__qualname__}<{self.name!r}; {self.type!r}>'



from __future__ import annotations

from typing import Iterator

from ..errors import EmptyPage
from ..post.page import *

__all__ = ('Search',)

_SEARCH_TEMPLATE = 'https://rezka.ag/search/?do=search&subaction=search&q=%s'


class Search(Page):
    """Ajax class for HDRezka search"""
    __slots__ = ('_query',)

    def __init__(self, query: str = ''):
        self.query = query.strip()
        super().__init__(_SEARCH_TEMPLATE % self.query)

    @property
    def query(self) -> str:
        return self._query

    @query.setter
    def query(self, value):
        self._query = value if isinstance(value, str) else str(value)
        self.page = _SEARCH_TEMPLATE % self.query

    @staticmethod
    def _concat_paginator(url: str):
        return f'{url}&page={{0}}'

    def __iter__(self) -> Iterator[InlineItem]:
        """
        Returns the generator of all found articles
        """
        page = 1
        while True:
            try:
                yield from self.page_iter(page)
            except EmptyPage:
                return
            page += 1

    def __repr__(self):
        return f"{self.__class__.__qualname__}({f'{self.query!r}' if self.query else ''})"



from __future__ import annotations

import setuptools

with open('requirements.txt', encoding='UTF-8') as f:
    install_requires = f.read().strip().splitlines()
with open('README.md', encoding='UTF-8') as f:
    long_description = f.read().strip()
with open('CHANGELOG.md', encoding='UTF-8') as f:
    long_description += f'\n\n---\n\n{f.read().strip()}\n'

setuptools.setup(
    name='HDRezka',
    version='1.1.1',

    author='Nikita (NIKDISSV)',
    author_email='nikdissv@proton.me',

    description='HDRezka (rezka.ag) Python API',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/NIKDISSV-Forever/HDRezka',

    packages=setuptools.find_packages(),
    install_requires=install_requires,

    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Multimedia :: Video',

    ],

    python_requires='>=3.9',
    keywords=['HDRezka', 'rezka.ag', 'watch online', 'api', 'stream']
)



from __future__ import annotations

import re
from base64 import b64decode

__all__ = ('clear_trash',)

_sub_trash = re.compile(
    '#h|//_//|I0A=|I0Ah|I0Aj|I0Ak|I0BA|I0Be|I14=|I14h|I14j|I14k|I15A|I15e|ISE=|ISEh|ISEj|ISEk|ISFA|ISFe|ISM=|ISMh'
    '|ISMj|ISMk|ISNA|ISNe|ISQ=|ISQh|ISQj|ISQk|ISRA|ISRe|IUA=|IUAh|IUAj|IUAk|IUBA|IUBe|IV4=|IV4h|IV4j|IV4k|IV5A|IV5e'
    '|IyE=|IyEh|IyEj|IyEk|IyFA|IyFe|IyM=|IyMh|IyMj|IyMk|IyNA|IyNe|IyQ=|IyQh|IyQj|IyQk|IyRA|IyRe|JCE=|JCEh|JCEj|JCEk'
    '|JCFA|JCFe|JCM=|JCMh|JCMj|JCMk|JCNA|JCNe|JCQ=|JCQh|JCQj|JCQk|JCRA|JCRe|JEA=|JEAh|JEAj|JEAk|JEBA|JEBe|JF4=|JF4h'
    '|JF4j|JF4k|JF5A|JF5e|QCE=|QCEh|QCEj|QCEk|QCFA|QCFe|QCM=|QCMh|QCMj|QCMk|QCNA|QCNe|QCQ=|QCQh|QCQj|QCQk|QCRA|QCRe'
    '|QEA=|QEAh|QEAj|QEAk|QEBA|QEBe|QF4=|QF4h|QF4j|QF4k|QF5A|QF5e|XiE=|XiEh|XiEj|XiEk|XiFA|XiFe|XiM=|XiMh|XiMj|XiMk'
    '|XiNA|XiNe|XiQ=|XiQh|XiQj|XiQk|XiRA|XiRe|XkA=|XkAh|XkAj|XkAk|XkBA|XkBe|Xl4=|Xl4h|Xl4j|Xl4k|Xl5A|Xl5e').sub


def clear_trash(trash_string: str) -> str:
    return b64decode(_sub_trash('', trash_string) + '==').decode()



from __future__ import annotations

import re
from inspect import getdoc
from typing import Any

from bs4 import Tag

from .fields import AgeRating, Duration, Place, Poster, Rating, Release
from ..urls import Quality
from ..._bs4 import BeautifulSoup

_find_any_digit = re.compile(r'\d+').findall


def _dummy(_):
    return NotImplemented


def _force_int(value: Any, /, default: int = 0) -> int:
    if not isinstance(value, str):
        value = str(value)
    if digits := _find_any_digit(value):
        return int(''.join(digits))
    return default


class PostInfo:
    __slots__ = ('rating', 'places', 'slogan', 'release', 'country', 'director', 'genre',
                 'quality', 'translator', 'age_rating', 'duration', 'from_', 'characters',
                 '_view', 'fields', 'title', 'orig_title', 'poster', 'description')
    FILL_FIELDS = __slots__[:13]

    def __init__(self, soup: BeautifulSoup):
        post_info = soup.find(class_='b-post__info')
        self.title = soup.find(class_='b-post__title').text.strip()
        orig_title = soup.find(class_='b-post__origtitle')
        self.orig_title = orig_title.text.strip() if orig_title else self.title
        poster = soup.select_one('.b-sidecover>a')
        self.poster = Poster(poster['href'], poster.find('img')['src'])
        self.description = soup.find(class_='b-post__description_text').text.strip()
        self._view = self._form_view(post_info.find_all('tr'))

        raw_fields = post_info.select('tr>td')
        if len(raw_fields) % 2:
            raw_fields.append(raw_fields[-1])

        fields = {raw_fields[i].find('h2').text.strip().removesuffix(':'): raw_fields[-~i]
                  for i in range(0, len(raw_fields), 2)}
        for field in self.FILL_FIELDS:
            meth = getattr(self, f'_get_field_{field}', _dummy)
            field_name = (getdoc(meth) or '').strip()
            item = fields.get(field_name)
            if not item:
                setattr(self, field, None)
                continue
            processed = meth(item)
            fields[field_name] = processed
            setattr(self, field.removeprefix('_get_field_'), processed)
        self.fields = fields

    @staticmethod
    def _get_field_rating(tag: Tag) -> dict[str, Rating]:
        """Рейтинги"""
        result = {}
        for span in tag.find_all('span', class_='b-post__info_rates'):
            classes: list[str] = span.get('class')
            classes.remove('b-post__info_rates')
            service = span.find('a').text
            votes = span.find('text')
            if votes:
                votes = int(votes.text.strip('()'))
            result[classes[0] if classes else service] = Rating(
                service=service, rating=float(span.find('span').text), votes=_force_int(votes))
        return result

    @staticmethod
    def _get_field_places(tag: Tag) -> tuple[Place]:
        """Входит в списки"""
        return *(Place(a.text, _force_int(a.next_element.next_element)) for a in tag.find_all('a')),

    @staticmethod
    def _get_field_slogan(tag: Tag) -> str:
        """Слоган"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_release(tag: Tag) -> Release:
        """Дата выхода"""
        return Release(day=tag.next_element.text.strip(), year=_force_int(tag.find('a').text))

    @staticmethod
    def _get_field_country(tag: Tag) -> tuple[str]:
        """Страна"""
        return *map(str.strip, tag.text.split(',')),

    @staticmethod
    def _get_field_director(tag: Tag) -> str:
        """Режиссер"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_genre(tag: Tag) -> tuple[str]:
        """Жанр"""
        return *map(str.strip, tag.text.split(',')),

    @staticmethod
    def _get_field_quality(tag: Tag):
        """В качестве"""
        return Quality(tag.text.strip())

    @staticmethod
    def _get_field_translator(tag: Tag) -> str:
        """В переводе"""
        return tag.get_text(strip=True)

    @staticmethod
    def _get_field_age_rating(tag: Tag) -> AgeRating:
        """Возраст"""
        return AgeRating(_force_int(span := tag.find('span').next_element), span.next_element.get_text(strip=True))

    @staticmethod
    def _get_field_duration(tag: Tag) -> Duration:
        """Время"""
        parts = tag.text.split(' ', 1) + [-1, '']
        return Duration(int(parts[0]), parts[1])

    @staticmethod
    def _get_field_from_(tag: Tag) -> tuple[str]:
        """Из серии"""
        return *(a.text for a in tag.find_all('a')),

    @staticmethod
    def _get_field_characters(tag: Tag) -> tuple[str]:
        """В ролях актеры"""
        return *(span.text for span in tag.select('span[itemprop="name"]')),

    def _form_view(self, table: list[Tag]) -> str:
        return '{0}\n\n{1}\n\n{2}'.format(
            self.title if self.orig_title == self.title else f'{self.title} ({self.orig_title})',
            '\n'.join(tr.text.replace(')', ') ').strip() for tr in table),
            self.description
        )

    def __str__(self):
        return self._view

    def __repr__(self):
        # noinspection PyUnboundLocalVariable
        return f"""{self.__class__.__qualname__}<{', '.join(
            f'{name}={attr!r}' for name in self.__slots__
            if not name.startswith('_') and (attr := getattr(self, name)))}>"""



from __future__ import annotations

from . import get_response
from ..errors import AjaxFail

__all__ = ('Ajax',)


class Ajax:
    """HDRezka Ajax class"""
    __slots__ = ()

    @classmethod
    def _send_data(cls, action: str, data: dict[str]) -> dict:
        answer: dict = get_response('POST', f'https://rezka.ag/ajax/{action}/', data=data).json()
        if not answer.get('success', True):
            raise AjaxFail(answer.get('message', answer))
        return answer

    @classmethod
    def get_cdn_series(cls, data: dict[str]):
        return cls._send_data('get_cdn_series', data)

    @classmethod
    def get_episodes(cls, id: int | str, translator_id: int | str):
        return cls.get_cdn_series(
            {'action': 'get_episodes',
             'id': id,
             'translator_id': translator_id})

    @classmethod
    def get_stream(cls, id: int | str, translator_id: int | str, season: int | str, episode: int | str):
        return cls.get_cdn_series(
            {'action': 'get_stream',
             'id': id,
             'translator_id': translator_id,
             'season': season,
             'episode': episode}
        )

    @classmethod
    def get_movie(cls, id: int | str, translator_id: int | str):
        return cls.get_cdn_series(
            {'action': 'get_movie',
             'id': id,
             'translator_id': translator_id}
        )
